/* Sort quizcard-entries for quizcard.c.
*/

extern char max_count;

#include "sort.h"

#include <stdio.h>                     /* fprintf() */
#include <string.h>		       /* strcmp() */
#include <stdlib.h>		       /* exit() */

static char *Wordform[8] = {"n", "v", "adj", "adv", "prep", "", "", ""};
static char  Gender[4]   = {'m', 'f', 'n'};
static char  Case[2]     = {'s', 'p'};

/* first we need a function which can compare two entries.  It will be
   qsort()-compatible, i.e. it returns:

   -1: first_entry < second_entry
    0: first_entry = second_entry
    1: first_entry > second_entry
*/

int compare(entry *first_entry, entry *second_entry)
{
  int res;

  res = strcmp(first_entry->phrase, second_entry->phrase);
  if (res == 0) {
    if (first_entry->byte1 < second_entry->byte1) {
      res = -1;
    } else if (first_entry->byte1 > second_entry->byte1) {
      res = 1;
    } /* res = 0 otherwise anyway... */
  }
  return res;
}

/* next we need to find a function which can find an entry in a sorted
   list of known length.  It will either return the index of the entry
   if found, or the index of the next-highest entry if not found.  An
   auxilliary flag signals which case we are dealing with... */

int find_entry(entry *the_entry, entry *sorted_entries[], int num_entries, byte *found)
{
  int ll;                              /* lower limit */
  int m;                               /* middle */
  int ul;                              /* upper limit */
  int res;
  
  /* special cases: */
  /* no entries: */
  if (num_entries == 0) {
    *found = FALSE;
    return 0;
  }
  /* in front of list: */
  if (compare(the_entry, sorted_entries[0]) < 0) {
    *found = FALSE;
    return 0;
  }
  /* after end of list: */
  if (compare(the_entry, sorted_entries[num_entries-1]) > 0) {
    *found = FALSE;
    return num_entries;
  }
  /* ok, so here we go: */
  ll = 0;
  ul = num_entries-1;
  while ((ul-ll) > 1) {
    m = (ul+ll)/2;
    res = compare(the_entry, sorted_entries[m]);
    if (res == 0) {
      *found = TRUE;
      return m;
    }
    if (res < 0) {
      ul = m;
    } else {
      ll = m;
    }
  } /* end while */
  /* better save then sorry: */
  res = compare(the_entry, sorted_entries[ll]);
  if (res == 0) {
    *found = TRUE;
    return ll;
  }
  if (res < 0) {
    fprintf(stderr, "find_entry(): This can't happen: entry < ll!\n");
  }
  res = compare(the_entry, sorted_entries[ul]);
  if (res == 0) {
    *found = TRUE;
    return ul;
  }
  if (res > 0) {
    fprintf(stderr, "find_entry(): This can't happen: entry > ul!\n");
  }
  *found = FALSE;
  return ul;
}

/* Ok, we obviously should be able to insert an entry into the list at
   a given position (and do not forget to increase num_entries by one
   afterwards!): */

int insert_entry(entry *the_entry, entry *sorted_entries[], int num_entries, int index)
{
  int i;

  /* Ok, here go a few special cases first: */
  /* list to small */
  if (index >= MAX_NUM_ENTRIES) {
    fprintf(stderr, "insert_entry(): Sorry, maximum number of entries (%d) exceeded...\n", MAX_NUM_ENTRIES);
    exit(1);
  }
  /* far beyond list... */
  if (index > num_entries) {
    fprintf(stderr, "insert_entry(): This can't happen: inserting far beyond list, n=%d, i=%d\n", num_entries, index);
    exit(1);
  }
  /* right at end of list */
  if (index == num_entries) {
    sorted_entries[index] = the_entry;
    return TRUE;
  }
  /* otherwise: */
  for (i=num_entries; i>index; i--) {
    sorted_entries[i] = sorted_entries[i-1];
  }
  sorted_entries[index] = the_entry;
  return TRUE;
}

/* build phrase --- memory for phrase needs to be allocated elsewhere... */
char *sprint_phrase(char* phrase, entry *the_entry)
{
  if (the_entry->phrase != NULL) {
    if (((the_entry->byte1 & WORDFORM) == NOUN) ||
        ((the_entry->byte1 & WORDFORM) == ADJECTIV)) {
      sprintf(phrase, "%s (%s%c%c)",
              the_entry->phrase,
              Wordform[(the_entry->byte1 & WORDFORM)>>4],
              Gender[(the_entry->byte1 & GENDER)>>2],
              Case[(the_entry->byte1 & CASE)>>1]);
    } else {
      sprintf(phrase, "%s (%s)",
              the_entry->phrase,
              Wordform[(the_entry->byte1 & WORDFORM)>>4]);
    } /* end if NOUN or ADJ. */
  } else {
    sprintf(phrase, "%s", "");
  }
  return phrase;
}

/* show all possible translations of an entry */
int show_translations(entry entries[], word index)
{
  int i;
  char phrase[MAX_PHRASE_LEN+1]=""; 

  printf("%s =\n", sprint_phrase(phrase, &entries[index]));
  for (i=0; i<entries[index].num_translations; i++) {
    printf("%d) %s\n",
           i+1,
           sprint_phrase(phrase, &entries[entries[index].translations[i]]));
  }
  return entries[index].num_translations;
}

/* show possible alternative translations */
int show_alternatives(entry entries[], word index, word leaveout,word numlines)
{
  int i;
  char phrase[MAX_PHRASE_LEN+1]="";
  int max_length;
  int length;
  int offset;
	char lead[40] = "                                       ";
	int lines = 1;
	char newline[3] = "\n";

  /* find longest entry */
  max_length = strlen(sprint_phrase(phrase, &entries[index]));
  for (i=0; i<entries[index].num_translations; i++) {
    if ((entries[index].translations[i] != leaveout) &&
	(entries[index].count[i] < max_count)) {
      lines++;
      length = strlen(sprint_phrase(phrase, &entries[entries[index].translations[i]]));
      if (length > max_length) {
	max_length = length;
      }				       /* end if length */
    }				       /* end if uncompleted */
  }				       /* end for i */
  if (max_length < 76) {
    offset = (80-max_length-3)/2;
  } else {
    offset = 0;
  }
  lead[offset] = 0;
  if (2*lines < numlines) {
    newline[0] = '\n';
    newline[1] = '\n';
    newline[2] = 0;
  } else {
    newline[0] = '\n';
    newline[1] = 0;
  }

  /* Print out translations: */
  printf("%s%s =%s\n", lead, sprint_phrase(phrase, &entries[index]), newline);
  for (i=0; i<entries[index].num_translations; i++) {
    if ((entries[index].translations[i] != leaveout) &&
	(entries[index].count[i] < max_count)) {
      printf("%s%d) %s%s",
	     lead,
	     entries[index].count[i],
	     sprint_phrase(phrase, &entries[entries[index].translations[i]]),
	     newline);
    }				       /* end if uncompleted */
  }				       /* end for i */
  return TRUE;
}
